@@ -25,6 +25,7 @@ struct node_t
         s_token = token_text;
         s_buf = lex_text;
         s_ws = ws_text;
+        s_attr = attr_text;
     }
     virtual ~node_t() {
         delete_children();
@@ -41,6 +42,7 @@ struct node_t
     string s_buf;
     string s_struct;
     string s_ws;
+    string s_attr;
     bool b_ignore;
     bool b_declaration_list;
     bool b_function_declaration;
@@ -73,8 +75,6 @@ struct node_t
         t_typeof,
         t_statement_list,
         t_compound_statement,
-        t_attribute,
-        t_attribute_list,
         t_string,
         t_string_list,
     } type;
@@ -173,6 +173,19 @@ struct node_t
         return false;
     }
 
+    bool identifiers_need_globalising(const string &func) const {
+      if (!s_identifier.empty())
+        if (is_global(s_identifier) || is_static_local(func, s_identifier))
+          return true;
+      
+      list<node_t *>::const_iterator ci = children.begin();
+      for(; ci != children.end(); ci++) {
+        if ((*ci)->identifiers_need_globalising(func))
+          return true;
+      }
+      return false;
+    }
+
     node_t *get_identifier() {
         if(s_identifier.size() > 0)
             return this;
@@ -226,6 +239,7 @@ struct node_t
         for(; ci != children.end(); ci++)
             (*ci)->print_to_str(dest);
 
+        // TODO: use an ostringstream for efficiency
         dest += s_buf;
     }
 
@@ -241,6 +255,8 @@ struct node_t
     bool handle_global(string func = "", list<string> *defined_globals = NULL) {
         if(get_function_declaration_b())
             return false;
+        if(!identifiers_need_globalising(func))
+            return false;
         
         string result;
         list<node_t *>::iterator ci = children.begin();
@@ -354,6 +370,8 @@ struct node_t
         string ws = get_whitespace(); remove_whitespace();
         string storage_class_specifiers;
 
+        string attr = declaration_specifiers->find_attr_and_remove();
+
         if(is_extern) {
             declaration_specifiers->remove_node("extern");
             storage_class_specifiers += "extern ";
@@ -371,17 +389,19 @@ struct node_t
             storage_class_specifiers += "register ";
         }
 
-        node_t *attr = NULL;
-        while( (attr = declaration_specifiers->find_attr_and_remove()) ) {
-            string a;
-            attr->print_to_str(a);
-            storage_class_specifiers += " ";
-            storage_class_specifiers += a;
-            delete attr;
+        if (declaration_specifiers->remove_node("__inline__") ||
+            declaration_specifiers->remove_node("__inline") ||
+            declaration_specifiers->remove_node("inline__")) {
+          storage_class_specifiers += "__inline__ ";
+        }
+
+        if (!attr.empty()) {
+            storage_class_specifiers += " "; // TODO: we might be adding more whitespace than we need here, check this
+            storage_class_specifiers += attr;
         }
         
         declaration_specifiers->reduce();
-        result = ws + "typedef " + declaration_specifiers->s_buf;
+        result = "typedef " + declaration_specifiers->s_buf;
         result += string(" ") + global_type_name + string("; ");
         // typedef done.
        
@@ -407,7 +427,7 @@ struct node_t
         list<node_t *>::iterator id_iter = init_declarators.begin();
         for(; id_iter != init_declarators.end(); ++id_iter) {
             node_t *id = *id_iter, *i, *pointer = NULL, *array = NULL,
-                *attribute = NULL, *initialiser_node = NULL;
+                *initialiser_node = NULL, *declarator = NULL;
             int stack;
             bool is_unbounded_array = false;
 
@@ -417,21 +437,11 @@ struct node_t
                 // declarator equals_t initializer
                 initialiser_node = id->children.back();
                 i = id->children.front();
-            } else if(id->type == node_t::t_init_declarator
-                    && id->children.size() == 4) {
-                // declarator attribute_specifier_list equals_t initializer
-                initialiser_node = id->children.back();
-                attribute = *(++id->children.begin());
-                i = id->children.front();
-            } else if(id->children.size() == 2 && 
-                    id->children.back()->type == t_attribute) {
-                // declarator attribute_specifier_list
-                attribute = id->children.back();
-                i = id->children.front();
             } else {
                 // declarator
                 i = id;
             }
+            declarator = i;
             // Iteratate past pointer nodes
             if(i->type == node_t::t_declarator) {
                 pointer = i->children.front();
@@ -485,62 +495,74 @@ struct node_t
 
             /* Special handling for section stuff; this should allow us to
              * globalise FreeBSD "linker sets" correctly. */
-            if(attribute && is_global_section(orig_iden)) {
-                node_t *sl = attribute->find_attribute_strings("__section__");
+            if (is_global_section(orig_iden)) {
+              node_t *attr_node = id->find_attribute_strings("__section__");
+              if (!attr_node) {
+                assert(storage_class_specifiers.find("__section__") != string::npos);
+              }
 
-                if(sl) {
-                    node_t *si = new node_t();
-                    sl->children.push_back(si);
+              if (!attr_node) {
+                assert(0); // TODO: handle this case by removing the section from the scs
+                // and putting it into each var
+                // At the moment we haven't implemented this case because it has
+                // never come up
+              }
 
-                    if(initialiser_node)
+              string orig_attr = attr_node->remove_attribute("__section__");
+              string iden_attrs = i->find_attr_and_remove();
+
+              if(initialiser_node)
                         initialiser_node->hgr_initialiser(func.empty() ? 
                             "__toplevel__undefined" : func);
 
-                    for(stack = 0; stack < NUM_STACKS; stack++) {
-                        result += type_name;
-                        if(pointer) pointer->print_to_str(result);
-                        result += string(" ");
+              for(stack = 0; stack < NUM_STACKS; stack++) {
+                result += type_name;
+                if(pointer) pointer->print_to_str(result);
+                result += string(" ");
 
-                        char stacknum[128];
-                        snprintf(stacknum, 128, "_global_section_%d_", stack);
+                char stacknum[128];
+                snprintf(stacknum, 128, "_global_section_%d_", stack);
 
-                        result += stacknum + orig_iden;
-        
-                        si->s_identifier = si->s_buf = string("\"") + stacknum
-                            + string("\"");
+                result += stacknum + orig_iden;
 
-                        attribute->print_to_str(result);
+                string new_attr = orig_attr;
+                
+                {
+                  string::size_type i, str_start, str_end;
 
-                        result += " = ";
+                  i = new_attr.find("__section__");
+                  i = new_attr.find("(", i+1); str_start = i + 1;
+                  i = new_attr.find(")", i+1); str_end = i;
 
-                        string init_str;
-                        initialiser_node->print_to_str(init_str);
+                  new_attr.insert(str_end, string(" \"") + stacknum + "\"");
+                }
 
-                        global_replace_toplevel(init_str, stack);
-                        global_old_replace_toplevel(init_str, stack);
+                result += " " + new_attr + " " + iden_attrs;
 
-                        result += init_str;
-                        result += ";";
-                    }
+                result += " = ";
+
+                string init_str;
+                initialiser_node->print_to_str(init_str);
+
+                global_replace_toplevel(init_str, stack);
+                global_old_replace_toplevel(init_str, stack);
+
+                result += init_str;
+                result += ";";
+              }
                     
-                    continue;
-                } else {
-                    fprintf(stderr, "Warning: Couldn't find __section__...\n");
-                    string sam;
-                    attribute->print_to_str(sam);
-                    fprintf(stderr, "attr= '%s'\n", sam.c_str());
-                }
-            } else if(attribute) {
-                extern int opt_verbose; 
-
-                if(opt_verbose > 0) {
-                    node_t *sl = attribute->find_attribute_strings("__section__");
-                    if(sl) {
-                        fprintf(stderr, "Warning: global '%s' has __section__ "
-                                "attribute but is not handled with $section.\n",
-                                orig_iden.c_str());
-                    }
+              continue;
+            } else {
+              extern int opt_verbose; 
+              if (opt_verbose > 0) {
+                node_t *attr_node = id->find_attribute_strings("__section__");
+                if (attr_node) {
+                  fprintf(stderr, "Warning: global '%s' has __section__ "
+                      "attribute but is not handled with $section. Attribute "
+                      "was: '%s'\n",
+                      orig_iden.c_str(), attr_node->s_attr.c_str());
                 }
+              }
             }
 
             if(is_global_section_startstop(orig_iden)) {
@@ -586,12 +608,23 @@ struct node_t
                 continue;
             }
 
-            if(!is_unbounded_array) {
+            string iden_attrs = i->find_attr_and_remove();
+
+            bool use_old_method = !is_unbounded_array;
+
+            if (global_expand_decl(orig_iden, func)) {
+              use_old_method = false;
+            }
+
+            if(use_old_method) {
                 // Use old method
                 string new_iden = global_old_name(orig_iden);
                 result += type_name;
                 if(pointer) pointer->print_to_str(result);
                 result += string(" ") + new_iden + string("[NUM_STACKS]");
+
+                if (!iden_attrs.empty())
+                  result += string(" ") + iden_attrs;
                 
                 // Need to do this before initialiser_node->hgr_initialiser()
                 global_define(orig_iden, new_iden, is_extern, is_volatile, 
@@ -607,9 +640,6 @@ struct node_t
 
                     string init_str;
                     initialiser_node->print_to_str(init_str);
-
-                    if(attribute)
-                        attribute->print_to_str(result);
                     
                     result += string(" = { ");
                     for(stack = 0; stack < NUM_STACKS; stack++) {
@@ -637,6 +667,9 @@ struct node_t
                 result += type_name;
                 i->s_buf = string(" ") + global_name(orig_iden, stack);
 
+                if (!iden_attrs.empty())
+                  result += string(" ") + iden_attrs;
+
                 // modify initialiser_node so it refers to the correct global
                 // instances.
                 if(initialiser_node)
@@ -657,7 +690,7 @@ struct node_t
         }
 
         delete_children();
-        s_buf = result;
+        s_buf = result + ws;
 
         return true;
     }
@@ -690,58 +723,6 @@ private:
         return result;
     }
 
-    node_t *find_attribute_strings(const string &attr_name) {
-        if(type == t_attribute && children.size() == 6) {
-            list<node_t *>::iterator ci = children.begin();
-            ++ci; // attribute_t -> left_bracket_t
-            ++ci; // left_bracket_t -> left_bracket_t
-            ++ci; // left_bracket_t -> attribute_list 
-
-            node_t *attr_list = *ci;
-            return attr_list->find_attribute_strings_attr_list(attr_name);
-        }
-        
-        list<node_t *>::iterator ci = children.begin();
-        for(; ci != children.end(); ++ci) {
-            node_t *ret = (*ci)->find_attribute_strings(attr_name);
-            if(ret)
-                return ret;
-        }
-        
-        return NULL;
-    }
-
-    node_t *find_attribute_strings_attr_list(const string &attr_name) {
-        node_t *attr = this;
-
-        if(type == t_attribute_list) {
-            attr = children.front()->find_attribute_strings_attr_list(attr_name);
-            if(attr) 
-                return attr;
-            else 
-                attr = children.back();
-        }
-
-        if(attr->children.size() != 4)
-            return NULL;
-
-        if(children.front()->s_identifier != attr_name)
-            return NULL;
-
-        // we are an attribute:
-        
-        list<node_t *>::iterator ci = children.begin();
-        ++ci; // identifier -> left_bracket_t
-        ++ci; // left_bracket_t -> expression
-
-        node_t *exp = *ci;
-
-        if(exp->type == t_string || exp->type == t_string_list)
-            return exp;
-
-        return NULL;
-    }
-
     string get_whitespace() {
         // the token is always at the end of the whitespace.
         string ws = s_ws;
@@ -753,27 +734,105 @@ private:
         return ws;
     }
 
-    node_t *find_attr_and_remove() {
-        list<node_t *>::iterator ci = children.begin();
-        for(; ci != children.end(); ++ci) {
-            node_t *c = *ci;
-            if(c->type == t_attribute) {
-                children.erase(ci);
-                return c;
-            }
-            c = c->find_attr_and_remove();
-            if(c)
-                return c;
-        }
-        return NULL;
-    }
-
     void remove_whitespace() {
         list<node_t *>::iterator ci = children.begin();
         for(; ci != children.end(); ++ci) {
             (*ci)->remove_whitespace();
         }
         s_buf = string(" ") + s_token;
+        if (!s_attr.empty()) {
+          s_buf += string(" ") + s_attr;
+        }
+    }
+
+    string find_attr_and_remove() {
+        std::ostringstream oss;
+        list<node_t *>::iterator ci = children.begin();
+        for(; ci != children.end(); ++ci) {
+          oss << (*ci)->find_attr_and_remove();
+        }
+        // We can't both have children with attrs, and have attrs ourself
+        // (only leafs can have attributes)
+        assert(oss.str().size() == 0 || s_attr.size() == 0);
+        if (s_attr.size() > 0) {
+          string::size_type s = s_buf.find(s_attr);
+          assert (s != string::npos);
+          s_buf.erase(s, s_attr.size());
+          string ret = s_attr;
+          s_attr.clear();
+          return ret + " ";
+        }
+        return oss.str();
+    }
+
+    string remove_attribute(const string& attr) {
+      if (s_attr.find(attr) != string::npos) {
+        string::size_type a_start, aa_start, i;
+        
+        a_start = s_attr.find(attr);
+        aa_start = s_attr.rfind("__attribute", a_start);
+
+        int depth = 0;
+        for (i = aa_start; i < s_attr.size(); i++) 
+        {
+          if (s_attr[i] == '(') depth++;
+          else if (s_attr[i] == ')') depth--;
+          if (i > a_start && depth == 0)
+            break;
+        }
+
+        string str = s_attr.substr(aa_start, i - aa_start);
+
+        string::size_type buf_i = s_buf.find(str);
+        assert(buf_i != string::npos);
+
+        s_buf.erase(buf_i, str.size());
+        s_attr.erase(aa_start, i - aa_start);
+
+        return str;
+      }
+
+      list<node_t *>::iterator ci = children.begin();
+      for(; ci != children.end(); ++ci) {
+        string s = (*ci)->remove_attribute(attr);
+        if (!s.empty())
+          return s;
+      }
+      return string();
+    }
+
+    bool has_attribute() {
+      if (!s_attr.empty())
+        return true;
+
+      list<node_t *>::iterator ci = children.begin();
+      for(; ci != children.end(); ++ci) {
+        if ((*ci)->has_attribute())
+          return true;
+      }
+      return false;
+    }
+    void print_attrs() {
+      list<node_t *>::iterator ci = children.begin();
+      for(; ci != children.end(); ++ci) {
+        (*ci)->print_attrs();
+      }
+      fprintf(stderr, "'%s':'%s'\n", s_buf.c_str(), s_attr.c_str());
+    }
+
+    node_t *find_attribute_strings(const string& s) {
+      if (!s_attr.empty()) {
+        if (s_attr.find(s) != string::npos) {
+          return this;
+        }
+      }
+      list<node_t *>::iterator ci = children.begin();
+      for(; ci != children.end(); ++ci) {
+        node_t* attr = (*ci)->find_attribute_strings(s);
+        if (attr)
+          return attr;
+      }
+      return NULL;
     }
     
     bool remove_node(const string &token)
