@@ -73,3 +73,6 @@ int main()
 }
 
 int get_stack_id() { return 0; }
+
+// This should generate a warning from the globaliser
+int notsection __attribute__ ((__section__ ("set_" "test"))) __attribute__((__used__)) = 0;
