@@ -547,13 +547,6 @@ __u32 secure_tcp_sequence_number(__u32 saddr, __u32 daddr,
     return 0x1020;
 }
 
-asmlinkage
-unsigned int csum_partial(const unsigned char * buff, int len, unsigned int sum)
-{
-    UNIMPLEMENTED_NOASSERT();
-    return 0;
-}
-
 void inode_init_once(struct inode *inode)
 {
     UNIMPLEMENTED_NOASSERT();
@@ -562,7 +555,7 @@ void inode_init_once(struct inode *inode)
 void iput(struct inode *inode)
 {
     UNIMPLEMENTED_NOASSERT();
-    // kfree ?
+    sock_destroy_inode(inode);
 }
 
 void fastcall prepare_to_wait(wait_queue_head_t *q, wait_queue_t *wait, int state)
@@ -634,17 +627,6 @@ void fastcall call_rcu_bh(struct rcu_head *head,
     func(head);
 }
 
-asmlinkage unsigned int  csum_partial_copy_generic(const unsigned char *src, unsigned char *dst,
-        int len, int sum, int *src_err_ptr, int *dst_err_ptr)
-{
-    if(dst_err_ptr)
-        *dst_err_ptr = 0;
-    
-    UNIMPLEMENTED_NOASSERT();
-    memcpy(dst, src, len);
-    return 0;
-}
-
 void fastcall
 prepare_to_wait_exclusive(wait_queue_head_t *q, wait_queue_t *wait, int state)
 {
