@@ -100,7 +100,7 @@ extern volatile unsigned long jiffies;
 
 void nsc_timer_tick(void)
 {
-    extern volatile unsigned long jiffies;
+  //    extern volatile unsigned long jiffies;
     struct pt_regs regs;
     memset(&regs, 0, sizeof(regs));
     //nsc_debugf("nsc_timer_tick\n");
@@ -261,7 +261,7 @@ void *nsc_socreate_tcp()
     nsc_assert(
             sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock)
             == 0, "sock_create failed in nsc_socreate_tcp");
-    sock->passcred = 0;
+    //sock->passcred = 0; /* EDIT: Fred for NSC: doesn't exist anymore in 2.6.14.2? */
 
     return sock;
 }
@@ -641,11 +641,12 @@ int nsc_do_sysctl(const char *sysctl_name, void *oldval, size_t *oldlenp,
             name[nlen++] = NET_IPV4;
         } else if(strncmp(s, "core/", 5) == 0) {
             name[nlen++] = NET_CORE;
-        } else if(strncmp(s, "tcp_westwood", 12) == 0) {
+/*      } else if(strncmp(s, "tcp_westwood", 12) == 0) {
             name[nlen++] = NET_TCP_WESTWOOD;
             // net/ipv4/tcp_westwood
             if(nlen != 3) return 1;
             break;
+*/ /* EDIT: Fred for NSC: No longer present in >= 2.6.13 */
         } else if(strncmp(s, "wmem_max", 8) == 0) {
             name[nlen++] = NET_CORE_WMEM_MAX;
             // net/core/wmem_max
@@ -656,16 +657,23 @@ int nsc_do_sysctl(const char *sysctl_name, void *oldval, size_t *oldlenp,
             // net/core/rmem_max
             if(nlen != 3) return 1;
             break;
-        } else if(strncmp(s, "rmem_default", 12) == 0) {
-            name[nlen++] = NET_CORE_RMEM_DEFAULT;
-            // net/core/rmem_default
+        // -- Fred NSC --
+	} else if(strncmp(s, "tcp_rmem", 8) == 0) {
+            name[nlen++] = NET_TCP_RMEM;
+            // net/ipv4/tcp_rmem
             if(nlen != 3) return 1;
             break;
-        } else if(strncmp(s, "wmem_default", 12) == 0) {
-            name[nlen++] = NET_CORE_WMEM_DEFAULT;
-            // net/core/wmem_default
+	} else if(strncmp(s, "tcp_wmem", 8) == 0) {
+            name[nlen++] = NET_TCP_WMEM;
+            // net/ipv4/tcp_wmem
             if(nlen != 3) return 1;
             break;
+	} else if(strncmp(s, "tcp_mem", 7) == 0) {
+            name[nlen++] = NET_TCP_MEM;
+            // net/ipv4/tcp_mem
+            if(nlen != 3) return 1;
+            break;
+	// --
         } else if(strncmp(s, "tcp_retrans_collapse", 18) == 0) {
             name[nlen++] = NET_IPV4_TCP_RETRANS_COLLAPSE;
             // net/ipv4/tcp_retrans_collapse
@@ -721,23 +729,20 @@ int nsc_do_sysctl(const char *sysctl_name, void *oldval, size_t *oldlenp,
             // net/ipv4/tcp_app_win
             if(nlen != 3) return 1;
             break;
-        } else if(strncmp(s, "tcp_rmem", 8) == 0) {
-            name[nlen++] = NET_TCP_RMEM;
-            // net/ipv4/tcp_rmem
+/* ------------------------------------------------------------- */
+	} else if(strncmp(s, "tcp_congestion_control", 22) == 0) {
+            name[nlen++] = NET_TCP_CONG_CONTROL;
+            // net/ipv4/tcp_congestion_control
             if(nlen != 3) return 1;
-            break;
-        } else if(strncmp(s, "tcp_wmem", 8) == 0) {
-            name[nlen++] = NET_TCP_WMEM;
-            // net/ipv4/tcp_wmem
-            if(nlen != 3) return 1;
-            break;
+	    break;
         }
+/* -- EDIT: Fred for NSC: new functionality in >= 2.6.13 --------*/
 
     } while( (s = strchr(s, '/')) );
 
     if(s == NULL)
         return 1;
-     
+
     return do_sysctl(name, nlen, oldval, oldlenp, newval, newlen);
 }
 
@@ -756,7 +761,7 @@ int nsc_setsockopt(void *so, char *optname, void *val, size_t valsize)
 int nsc_get_tcp_var(void *so, const char *var, char *result, int rlen)
 {
     struct sock *sock = ((struct socket *)so)->sk;
-    struct tcp_opt *tp = tcp_sk(sock);
+    struct tcp_sock *tp = tcp_sk(sock);
 
          if(strcmp(var, "srtt_") == 0)  // in ticks
     {
@@ -794,7 +799,7 @@ int nsc_get_tcp_var(void *so, const char *var, char *result, int rlen)
     }
     else if(strcmp(var, "rxtcur_") == 0)
     {
-        snprintf(result, rlen, "%f", (float)tp->rto);
+        snprintf(result, rlen, "%f", (float)tp->inet_conn.icsk_rto); /* EDIT: Fred for NSC: changed in 2.6.14.2 (inet_conn) */
         return 1;
     }
 
@@ -830,8 +835,7 @@ void nsc_sysctl_r(struct proc_dir_entry *p, const char *d)
         if(table->proc_handler)
         {
             char rbuf[6000];
-            size_t len = 5000;
-            loff_t pos = 0;
+            //unsigned int len = 5000;
             struct file f;
             f.f_pos = 0;
 
@@ -839,7 +843,7 @@ void nsc_sysctl_r(struct proc_dir_entry *p, const char *d)
             memset(rbuf, 0, 5000);
             
             // TODO: make this work for 2.6, below is for 2.4
-            table->proc_handler(table, 0, &f, rbuf, &len, &pos);
+            //table->proc_handler(table, 0, &f, rbuf, &len);
 
             if(strlen(rbuf)) {
                 printk("%s = ", buf2);
