@@ -19,7 +19,7 @@
   Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
 
-  $Id: nsc.cc 1294 2006-04-20 15:25:01Z stj2 $
+  $Id: nsc.cc 1665 2008-01-10 20:42:26Z stj2 $
   
   */
 
@@ -35,6 +35,10 @@
 #include <dlfcn.h>
 #include <stdlib.h>
 
+#include <sstream>
+#include <string>
+#include <iterator>
+
 extern "C" {
 #include <pcap.h>
 }
@@ -49,6 +53,7 @@ extern "C" {
 #include "node.h"
 
 //#define DBG_MSGS 1
+#define UNUSED __attribute__((unused))
 
 /* functions from the stack and supporting code... These are mostly legacy
  * and aren't really used anymore. */
@@ -97,7 +102,7 @@ NSCAgent::NSCAgent(const char *libname)
     soft_timer_(this), wakeup_timer_(this), stack_(NULL), con_(NULL), 
     con2_(NULL), to_snd_(0), packet_size_(1500), wnd_size_(20), 
     packet_dump_enabled_(false), has_received_(false), close_on_empty_(false),
-    send_type_(InfiniteBuffers), bytes_sent_(0), bytes_recv_(0), pdump_(NULL), 
+    send_type_(NoBuffers), bytes_sent_(0), bytes_recv_(0), pdump_(NULL), 
     pcap_(NULL)
 {
     bind("bytes_", &bytes_recv_);
@@ -220,7 +225,7 @@ void NSCAgent::sendlow(int datalen)
 {
     /* Send to stack here, it should reach us via a callback function 
      * later which will be called by the network stack code. */
-    int packetsize = datalen;
+    size_t packetsize = datalen;
     INetStreamSocket *sock = NULL;
     const size_t buf_size = 65535;
     static unsigned char buf[buf_size];
@@ -317,7 +322,7 @@ void NSCAgent::sendlow(int datalen)
 }
 
 /** Dumps a packet out to (pcap) file. This is for making packet traces. */
-void NSCAgent::dump(void *d, int size)
+void NSCAgent::dump(const void *d, int size)
 {
     const int snaplen = 400;
 
@@ -387,6 +392,9 @@ void NSCAgent::recv(Packet* pkt, Handler*)
     if(to_snd_ > 0) {
         sendlow(to_snd_ > packet_size_ ? packet_size_ : to_snd_);
     }
+
+    // haxor, want to check this:
+    readlow();
 }
 
 bool NSCAgent::get_connection_var(const char *var, char *buf, int buflen)
@@ -507,7 +515,7 @@ int NSCAgent::command(int argc, const char*const* argv)
         if (strcmp(argv[1], "nsc_set") == 0) {
             fprintf(stderr, "NSC: %s %s %s %s\n",
                     argv[0],argv[1],argv[2],argv[3]);
-            if(strcmp(argv[2], "window_") == 0) {
+	      if(strcmp(argv[2], "window_") == 0) {
                 int val = atoi(argv[3]), error = 0;
 
                 val *= packet_size_ - 40;
@@ -583,15 +591,19 @@ int NSCAgent::command(int argc, const char*const* argv)
                         sizeof(newval));
 
                 // add debugging info
-                /* printf(stderr, "sysctl: %s, oldval=%s, newval=%s, error=%d\n",
-                argv[2], oldval, newval, error); */
+                 printf("sysctl: %s, oldval=%s, newval=%s, error=%d\n",
+			argv[2], oldval, newval, error); 
 
                 assert(error == 0);
 
                 return TCL_OK;
             }
             
-            int val = atoi(argv[3]), oldval;
+            int val, oldval;
+	    if (strncmp(argv[3], "0x", 2) == 0)
+	      val = strtoll(argv[3]+2, (char **)NULL, 16);
+	    else
+	      val = atoi(argv[3]);
             size_t oldvalsize = sizeof(int);
             int error __attribute__((__unused__)) =
             stack_->sysctl(argv[2], &oldval, &oldvalsize, &val, sizeof(val));
@@ -611,6 +623,22 @@ int NSCAgent::command(int argc, const char*const* argv)
             add_interface(argv[2], argv[3]); 
             return TCL_OK;
         }
+	else if (strcmp(argv[1], "sysctl") == 0) {
+           unsigned int val[2], oldval[2];
+	    if (strncmp(argv[3], "0x", 2) == 0)
+	      val[0] = strtoll(argv[3]+2, (char **)NULL, 16);
+	    else
+	      val[0] = atoi(argv[3]);
+	    if (strncmp(argv[4], "0x", 2) == 0)
+	      val[1] = strtoll(argv[4]+2, (char **)NULL, 16);
+	    else
+	      val[1] = atoi(argv[4]);
+            size_t oldvalsize = sizeof(oldval);
+            int error __attribute__((__unused__)) =
+            stack_->sysctl(argv[2], oldval, &oldvalsize, val, sizeof(val));
+            assert(error == 0);
+            return TCL_OK;	  
+	}
     } else if (argc == 6) {
         /* Another hack to support other sysctls. This interface really needs to be
            updated so we no longer need to go this. */
@@ -643,7 +671,7 @@ int NSCAgent::command(int argc, const char*const* argv)
                    %d",atoi(argv[3]),atoi(argv[4]),atoi(argv[5]));
                    */
 
-                int error = stack_->sysctl(argv[2], oldval, &oldvalsize, newval,
+                int UNUSED error = stack_->sysctl(argv[2], oldval, &oldvalsize, newval,
                         sizeof(newval));
 
                 // add debugging info
@@ -657,6 +685,18 @@ int NSCAgent::command(int argc, const char*const* argv)
                 return TCL_OK;
             }
         }
+    } 
+    
+    if (argc > 2 && strcmp(argv[1], "stack_cmd") == 0) {
+      // Allows sending arbitrary text commands to a stack
+      std::ostringstream oss;
+      std::copy(argv + 2, argv + argc, std::ostream_iterator<const char *>(oss, " "));
+      std::string str = oss.str();
+      str = str.substr(0, str.size() - 1);
+
+      stack_->cmd(str.c_str());
+
+      return TCL_OK;
     }
     ret = (Agent::command(argc, argv));
 #ifdef DBG_MSGS
@@ -667,10 +707,10 @@ int NSCAgent::command(int argc, const char*const* argv)
 
 void NSCAgent::gettime(unsigned int *sec, unsigned int *usec)
 {
-    float time = Scheduler::instance().clock();
+    double time = Scheduler::instance().clock();
 
     *sec = (unsigned int)time;
-    *usec = (unsigned int)(time * 1000000.0f);
+    *usec = (unsigned int)((time - (double)*sec) * 1000000.0f);
 }
 
 /** This is called once every tick, and is responsible for calling the
@@ -751,7 +791,6 @@ void NSCAgent::readlow()
                         buflen);
 #endif
             }
-
             if(buflen && app_ && !err) {
 #ifdef DBG_MSGS
                 fprintf(stderr, "%f app recv %d bytes\n",
@@ -793,8 +832,9 @@ void NSCAgent::readlow()
     }
 }
 
+
 /** Called from the stack when a packet needs to be output to the wire. */
-void NSCAgent::send_callback(void *data, int datalen)
+void NSCAgent::send_callback(const void *data, int datalen)
 {
     Packet* pkt = allocpkt();
     PacketData *pdata = new PacketData(datalen);
@@ -941,16 +981,18 @@ void NSCAgent::send_callback(void *data, int datalen)
 
     // Send to lower level (link layer or whatever)
     target_->recv(pkt);
+
+    // New method: if_send_finish. This should be called by the nic
+    // driver code when there is space in it's queue. Unfortunately
+    // we don't know whether that is the case from here, so we just
+    // call it immediately for now. **THIS IS INCORRECT**. It should
+    // only be called when a space becomes free in the nic queue.
+    stack_->if_send_finish(0); // XXX: hardcoded if id
+
     // XXX: Do we need this?
     idle();
 }
 
-/** Just calls software_interrupt. Not sure why I have this method. */
-void NSCAgent::interrupt()
-{
-    software_interrupt();
-}
-
 /** This is called when something interesting happened to a socket. Perhaps
  * a socket can now be written to or read from or something. We cannot
  * directly do things like read or write from the socket, because this is
@@ -965,7 +1007,9 @@ void NSCAgent::interrupt()
 void NSCAgent::wakeup()
 {
     // schedule software_interrupt + sendmsg
-    wakeup_timer_.resched(0.0005);
+    if (wakeup_timer_.status() == TIMER_IDLE) {
+        wakeup_timer_.resched(0.000001); // 1us
+    }
 }
 
 /** When we get a wakeup we know something interesting has happened to a
@@ -1002,12 +1046,7 @@ void NSCAgent::add_interface(const char *a, const char *b)
 
 void NSCAgent::set_default_gw(const char *dest)
 {
-    unsigned int a1,a2,a3,a4;
-
-    sscanf(dest, "%d.%d.%d.%d", &a1,&a2,&a3,&a4);
-    stack_->add_default_gateway(
-            (unsigned char)a1, (unsigned char)a2,
-            (unsigned char)a3, (unsigned char)a4);
+    stack_->add_default_gateway(dest);
 }
 
 void NSCAgent::tcp_connect(const char *a, int b)
@@ -1025,53 +1064,28 @@ void NSCAgent::tcp_listen(int port)
     con_->listen(port);
 }
 // -----------------------------------------------------------------------
-static class Linux24SimpleAgentClass : public TclClass {
-    public:
-        Linux24SimpleAgentClass() : TclClass("Agent/TCP/NSC/Linux24") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("liblinux24.so"));
-        }
-} class_linux24_simple_agent;
 
-static class Linux26SimpleAgentClass : public TclClass {
-    public:
-        Linux26SimpleAgentClass() : TclClass("Agent/TCP/NSC/Linux26") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("liblinux26.so"));
-        }
-} class_linux26_simple_agent;
+#define REGISTER_SIMPLE_NSC_TCP_AGENT(name,filename) \
+  static class NscTcpSimpleAgentClass_##name : public TclClass { \
+      public: \
+          NscTcpSimpleAgentClass_##name() : TclClass("Agent/TCP/NSC/"#name) {} \
+          TclObject* create(int, const char*const*) { \
+              return (new NSCSimpleAgent(filename)); \
+          } \
+  } class_nsc_tcp_simple_agent_##name
 
-static class Linux26142SimpleAgentClass : public TclClass {
-    public:
-        Linux26142SimpleAgentClass() : TclClass("Agent/TCP/NSC/Linux26142") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("liblinux2.6.14.2.so"));
-        }
-} class_linux26142_simple_agent;
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux24,        "liblinux2.4.28.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux26,        "liblinux2.6.10.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux26142,     "liblinux2.6.14.2.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux2618,      "liblinux2.6.18.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux2619,      "liblinux2.6.19.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux2620,      "liblinux2.6.20.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux2621,      "liblinux2.6.21.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(Linux2622,      "liblinux2.6.22.so");
 
-static class lwipSimpleAgentClass : public TclClass {
-    public:
-        lwipSimpleAgentClass() : TclClass("Agent/TCP/NSC/lwIP") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("liblwip.so"));
-        }
-} class_lwip_simple_agent;
-
-static class FreeBSD5SimpleAgentClass : public TclClass {
-    public:
-        FreeBSD5SimpleAgentClass() : TclClass("Agent/TCP/NSC/FreeBSD5") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("libfreebsd5.so"));
-        }
-} class_freebsd5_simple_agent;
-
-static class OpenBSD3SimpleAgentClass : public TclClass {
-    public:
-        OpenBSD3SimpleAgentClass() : TclClass("Agent/TCP/NSC/OpenBSD3") {}
-        TclObject* create(int, const char*const*) {
-            return (new NSCSimpleAgent("libopenbsd3.so"));
-        }
-} class_openbsd3_simple_agent;
+REGISTER_SIMPLE_NSC_TCP_AGENT(lwIP,           "liblwip.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(FreeBSD5,       "libfreebsd5.3.so");
+REGISTER_SIMPLE_NSC_TCP_AGENT(OpenBSD3,       "libopenbsd3.5.so");
 
 static class GenericNSCSimpleAgentClass : public TclClass {
     public:
@@ -1271,6 +1285,7 @@ int NSCSimpleAgent::command(int argc, const char*const* argv)
             return TCL_OK;
         }
     }
+
     return NSCAgent::command(argc, argv);
 }
 
